import com.haxepunk.HXP;
import com.haxepunk.graphics.Tilemap;
import com.haxepunk.masks.Grid;
import haxe.xml.Fast;

import flash.utils.ByteArray;

class TiledSet {
	public var first: Int;
	public var count: Int;
	public var image:String;
	public function new(i:String, f:Int, c:Int) { image=i; first=f; count=c; }
}

class TiledLayer {
	public var data: Array<Int>;
	public function new() {}
}

class TiledObject {
	public var x: Float;
	public var y: Float;
	public var width: Float;
	public var height: Float;
	public var type:String;
	public var name:String;
	public var properties: Map<String, String>;
	public function new(type:String, name:String, x:Float, y:Float) {
		this.x = x;
		this.y = y;
		this.name = name;
		this.type = type;
		width = height = 0;
	}
}


/** More generic Tiled level parser */
class Tiled {

	public var width (default,null): Int;
	public var height (default,null): Int;

	public var tileWidth (default,null): Int;
	public var tileHeight (default,null): Int;


	public var properties (default,null): Map<String, String>;


	private var tilesets: Map< String, TiledSet >;
	private var layers:   Map< String, TiledLayer >;
	private var objects:  Map< String, Array<TiledObject> >;

	public function new(file:String) {
		var data:String = openfl.Assets.getText(file);
		loadLevel( new Fast( Xml.parse(data) ) );
	}

	public function isLoaded() { return width>0 && height>0; }
	

	/** Get a tile index
	 * @param name	Tile layer name
	 * @param x		x coordinate of tile
	 * @param y		y coordinate of tile
	 * @param tileset Optional tileset name
	 */
	public function tile(name:String, x:Int, y:Int, tileset:String=null): Int {
		if(layers.exists(name)) {
			var index:Int = layers[name].data[ x + y*width ];
			if(tileset!=null && tilesets.exists(tileset)) {
				index -= tilesets[tileset].first;
				if(index<0 || index>=tilesets[tileset].count) return -1;
			}
			return index;
		} else return -1;
	}



	/** Create a tilemap from a layer and tileset
	 * @param layer The layer name
	 * @param tileset The name of the tileset
	 * @param tileImage Overrides tile image filename found in file
	 */
	public function createTilemap(layer:String, tileset:String, ?tileImage:String=null): Tilemap {
		if(!layers.exists(layer)) throw("Layer " + layer + " not found in map");
		if(!tilesets.exists(tileset)) throw("Tileset " + tileset + " not found in map");
		if(tileImage==null) tileImage = tilesets[tileset].image;
		var map:Tilemap = new Tilemap(tileImage, width*tileWidth, height*tileHeight, tileWidth, tileHeight);
		var a:TiledLayer = layers[layer];
		var offset:Int=1, count:Int=64, used:Int=0;
		if(tilesets.exists(tileset)) {
			offset = tilesets[tileset].first;
			count = tilesets[tileset].count;
		} else for(ts in tilesets) {
			offset = ts.first;
			count = ts.count;
			break;
		}
		//trace(tileset + " Offset: " + offset + " Count: " + count);
		for(i in 0...a.data.length) {
			var tile:Int = a.data[i] - offset;
			if(tile>=0 && tile<count) {
				map.setTile( i%width, Math.floor(i/width), tile);
				used++;
			}
		}
		return used>0? map: null;
	}

	/** Create a clipmap grid from a layer and tile index
	 * @param name The layer name
	 * @param tileset The name of the tileset to use
	 * @param type The index of the tile in the tileset to be the 'on' state
	 */
	public function createClipmap(name:String, tileset:String, solid:Dynamic): Grid {
		if(!layers.exists(name)) return null;
		var grid:Grid = new Grid(width*tileWidth, height*tileHeight, tileWidth, tileHeight);
		var a:TiledLayer = layers[name];
		var offset:Int = tilesets.exists(tileset)? tilesets[tileset].first: 1;
		var tiles: Int=0, s:Bool;
		for(i in 0...a.data.length) {
			if(Std.is(solid, Int)) s = a.data[i] - offset == solid;
			else if(Std.is(solid, Array)) s = Lambda.indexOf(solid, a.data[i]-offset) >= 0;
			else s = false;
			grid.setTile( i%width, Math.floor(i/width), s);
			if(s) ++tiles;
		}
		return tiles>0? grid: null; //Null if no collision
	}

	/** Get an array of objects */
	public function getObjects(type:String=null, group:String=null): Array<TiledObject> {
		var list:Array<TiledObject> = new Array<TiledObject>();
		if(group!=null) appendObjects(list, objects[group], type);
		else for(g in objects) appendObjects(list, g, type);
		return list;
	}
	public function appendObjects(out:Array<TiledObject>, list:Array<TiledObject>, type:String=null) {
		for(o in list) {
			if(type==null || type==o.type) out.push(o);
		}
	}



	//// //// LOADING //// ////

	private function loadLevel(data:Fast): Bool {
		try {
			data = data.node.map;
			var item:Fast;


			tileWidth  = Std.parseInt( data.att.tilewidth );
			tileHeight = Std.parseInt( data.att.tileheight );
			width  = Std.parseInt( data.att.width );
			height = Std.parseInt( data.att.height );

			//Properties
			properties = readProperties(data);


			//Tilesets
			tilesets = new Map<String, TiledSet>();
			for (item in data.nodes.tileset) {
				var iw: Int = Std.parseInt( item.node.image.att.width );
				var ih: Int = Std.parseInt( item.node.image.att.height );
				var tw: Int = Std.parseInt( item.att.tilewidth );
				var th: Int = Std.parseInt( item.att.tileheight );
				var first:Int = Std.parseInt( item.att.firstgid );
				var count:Int = Std.int( (iw/tw) * (ih/th) );
				tilesets [ item.att.name ] = new TiledSet( item.node.image.att.source, first, count );
			}

			// Layers
			layers = new Map<String, TiledLayer>();
			for (item in data.nodes.layer) {
				var layer: TiledLayer = loadLayer(item);
				if(layer!=null) layers[ item.att.name ] = layer;
			}

			// Objects
			objects = new Map<String, Array<TiledObject> >();
			for(group in data.nodes.objectgroup) {
				var list:Array<TiledObject> = objects[group.att.name] = new Array<TiledObject>();
				for(item in group.nodes.object) {
					var px: Float = Std.parseFloat( item.att.x );
					var py: Float = Std.parseFloat( item.att.y );
					var type: String = item.has.type? item.att.type: "";
					var name: String = item.has.name? item.att.name: "";
					var object = new TiledObject( type, name, px, py);
					if(item.has.width) object.width = Std.parseFloat( item.att.width );
					if(item.has.height) object.height = Std.parseFloat( item.att.height );
					object.properties = readProperties( item );
					list.push(object);
				}
			}
		}
		catch(e:String) {
			trace(e);
			return false;
		}
		return true;
	}

	private function readProperties(s:Fast): Map<String, String> {
		if(!s.hasNode.properties || !s.node.properties.hasNode.property) return null;
		var propList = new Map<String, String>();
		for (node in s.node.propertes.nodes.property) {
			for(prop in s.nodes.property) {
				propList[ node.att.name ] = node.att.value;
			}
		}
		return propList;
	}

	private function loadLayer(src:Fast): TiledLayer {
		var w:Int = Std.parseInt( src.att.width );
		var h:Int = Std.parseInt( src.att.height );
		var layer: TiledLayer = new TiledLayer();
		//Encoding type
		var data:String = StringTools.trim( src.node.data.innerData );
		var encoding:String = src.node.data.att.encoding; //For some reason, switching directly fails
		var compression:String = src.node.data.att.compression;
		switch(encoding) {
		case "base64":
			//Compression?
			switch(compression) {
			case "none":
				layer.data = makeArray( decode64(data) );
			case "zlib":
				var tmp:ByteArray = decode64(data);
				tmp.uncompress();
				layer.data = makeArray(tmp);
			default:
				throw "Cannot decode compression: '" +src.node.data.att.compression + "'";
			}
		case "csv":
			var strings:Array<String> = data.split(',');
			layer.data = new Array<Int>();
			for(s in strings) layer.data.push( Std.parseInt(s) );
		default:
			throw "Cannot decode: '" + src.node.data.att.encoding + "'";
		}
		return layer;
	}

	private static var BASE64:String = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
	private function decode64(data:String): ByteArray {
		// Create lookup array
		var lookup: Map<Int, Int> = new Map<Int, Int>();
		for(c in 0 ... 65) lookup[ BASE64.charCodeAt(c) ] = c;
		// Read data: 4 characters => 1 byte
		var i:Int = 0;
		var buffer:ByteArray = new ByteArray();
		while(i<data.length) {
			var a0:Int = lookup[ data.charCodeAt(i+0) ];
			var a1:Int = lookup[ data.charCodeAt(i+1) ];
			var a2:Int = lookup[ data.charCodeAt(i+2) ];
			var a3:Int = lookup[ data.charCodeAt(i+3) ];

			buffer.writeByte( (a0 << 2) | ((a1&0x30) >> 4) );
			buffer.writeByte(((a1&0x0f) << 4) | ((a2&0x3c) >> 2));
			buffer.writeByte(((a2&0x03) << 6) | a3);
			i += 4;
		}
		buffer.position = 0;
		return buffer;
	}
	private function makeArray(raw:ByteArray): Array<Int> {
		raw.endian = flash.utils.Endian.LITTLE_ENDIAN; //"littleEndian";
		var out:Array<Int> = new Array<Int>();
		while(raw.position < raw.length) out.push( raw.readInt() );
		return out;
	}
	

}

